
# Shape of the function
Convex = "convex"
"""
Convex: convex function.
"""

Concave = "concave"
"""
Concave: concave funciton.
"""

FUN_Categories = {
    Convex: "convex function",
    Concave: "concave funciton"
}

# Optimization
OPT_LOCAL = "local"
OPT_DEFAULT = None
